import pandas as pd
import numpy as np


def annotation_chunker(kaleidoscope_df, chunk_length):
    """
    Function that converts a Kaleidoscope-formatted Dataframe containing 
    annotations to uniform chunks of chunk_length.

    Note: if all or part of an annotation covers the last < chunk_length
    seconds of a clip it will be ignored. If two annotations overlap in 
    the same 3 second chunk, both are represented in that chunk

    Args:
        kaleidoscope_df (Dataframe)
            - Dataframe of annotations in kaleidoscope format

        chunk_length (int)
            - duration to set all annotation chunks
    Returns:
        Dataframe of labels with chunk_length duration 
        (elements in "OFFSET" are divisible by chunk_length).
    """

    #Init list of clips to cycle through and output dataframe
    clips = kaleidoscope_df["IN FILE"].unique()
    df_columns = {'IN FILE' :'str', 'CLIP LENGTH' : 'float64', 'CHANNEL' : 'int64', 'OFFSET' : 'float64',
                'DURATION' : 'float64', 'SAMPLE RATE' : 'int64','MANUAL ID' : 'str'}
    output_df = pd.DataFrame({c: pd.Series(dtype=t) for c, t in df_columns.items()})
    
    # going through each clip
    for clip in clips:
        clip_df = kaleidoscope_df[kaleidoscope_df["IN FILE"] == clip]
        birds = clip_df["MANUAL ID"].unique()
        sr = clip_df["SAMPLE RATE"].unique()[0]
        clip_len = clip_df["CLIP LENGTH"].unique()[0]

        # quick data sanitization to remove very short clips
        # do not consider any chunk that is less than chunk_length
        if clip_len < chunk_length:
            continue
        potential_annotation_count = int(clip_len)//int(chunk_length)

        # going through each species that was ID'ed in the clip
        arr_len = int(clip_len*1000)
        for bird in birds:
            species_df = clip_df[clip_df["MANUAL ID"] == bird]
            human_arr = np.zeros((arr_len))
            # looping through each annotation
            for annotation in species_df.index:
                minval = int(round(species_df["OFFSET"][annotation] * 1000, 0))
                # Determining the end of a human label
                maxval = int(
                    round(
                        (species_df["OFFSET"][annotation] +
                         species_df["DURATION"][annotation]) *
                        1000,
                        0))
                # Placing the label relative to the clip
                human_arr[minval:maxval] = 1
            # performing the chunk isolation technique on the human array

            for index in range(potential_annotation_count):
                chunk_start = index * (chunk_length*1000)
                chunk_end = min((index+1)*chunk_length*1000,arr_len)
                chunk = human_arr[int(chunk_start):int(chunk_end)]
                if max(chunk) >= 0.5:
                    row = pd.DataFrame(index = [0])
                    annotation_start = chunk_start / 1000
                    #updating the dictionary
                    row["IN FILE"] = clip
                    row["CLIP LENGTH"] = clip_len
                    row["OFFSET"] = annotation_start
                    row["DURATION"] = chunk_length
                    row["SAMPLE RATE"] = sr
                    row["MANUAL ID"] = bird
                    row["CHANNEL"] = 0
                    output_df = pd.concat([output_df,row], ignore_index=True)
    return output_df


from .annotation_post_processing import annotation_chunker
from .IsoAutio import build_isolation_parameters_microfaune, generate_automated_labels, kaleidoscope_conversion, isolate
from .visualizations import annotation_duration_histogram, binary_visualization, local_line_graph, spectrogram_graph, spectrogram_visualization
from .statistics import annotation_duration_statistics, class_statistics, automated_labeling_statistics, clip_general, clip_IoU, clip_statistics, global_dataset_statistics, global_statistics, matrix_IoU_Scores


import pandas as pd
from scipy import stats
import numpy as np
import time

# Function that takes in a pandas dataframe of annotations and outputs a
# dataframe of the mean, median, mode, quartiles, and standard deviation of
# the annotation durations.
def annotation_duration_statistics(df):
    """
    Function that calculates basic statistics related to the duration of
    annotations of a Pandas Dataframe compatible with PyHa.

    Args:
        df (Pandas Dataframe)
            - Automated labels or manual labels.

    Returns:
        Pandas Dataframe containing count, mean, mode, standard deviation, and
        IQR values based on annotation duration.
    """
    # Reading in the Duration column of the passed in dataframe as a Python
    # list
    annotation_lengths = df["DURATION"].to_list()
    # converting to numpy array which has more readily available statistics
    # functions
    annotation_lengths = np.asarray(annotation_lengths)
    # Converting the Python list to a numpy array
    entry = {'COUNT': np.shape(annotation_lengths)[0],
             'MODE': stats.mode(np.round(annotation_lengths, 2))[0][0],
             'MEAN': np.mean(annotation_lengths),
             'STANDARD DEVIATION': np.std(annotation_lengths),
             'MIN': np.amin(annotation_lengths),
             'Q1': np.percentile(annotation_lengths, 25),
             'MEDIAN': np.median(annotation_lengths),
             'Q3': np.percentile(annotation_lengths, 75),
             'MAX': np.amax(annotation_lengths)}
    # returning the dictionary as a pandas dataframe
    return pd.DataFrame.from_dict([entry])


def clip_general(automated_df, human_df):
    """
    Function to generate a dataframe with statistics relating to the efficacy
    of the automated label compared to the human label.
    These statistics include true positive, false positive, false negative,
    true negative, union, precision, recall, F1, and Global IoU. For general
    clip overlap

    Args:
        automated_df (Dataframe)
            - Dataframe of automated labels for one clip

        human_df (Dataframe)
            - Dataframe of human labels for one clip.

    Returns:
        Dataframe with general clip overlap statistics comparing the automated
        and human labeling.
    """
    # This looks at one class across one clip
    clip_class = human_df["MANUAL ID"]
    clip_class = list(dict.fromkeys(clip_class))[0]
    duration = automated_df["CLIP LENGTH"].to_list()[0]
    SAMPLE_RATE = automated_df["SAMPLE RATE"].to_list()[0]
    # Initializing two arrays that will represent the human labels and
    # automated labels with respect to the audio clip
    # print(SIGNAL.shape)
    human_arr = np.zeros((int(SAMPLE_RATE * duration),))
    bot_arr = np.zeros((int(SAMPLE_RATE * duration),))
    
    folder_name = automated_df["FOLDER"].to_list()[0]
    clip_name = automated_df["IN FILE"].to_list()[0]
    # Placing 1s wherever the au
    for row in automated_df.index:
        minval = int(round(automated_df["OFFSET"][row] * SAMPLE_RATE, 0))
        maxval = int(
            round(
                (automated_df["OFFSET"][row] +
                 automated_df["DURATION"][row]) *
                SAMPLE_RATE,
                0))
        bot_arr[minval:maxval] = 1
    for row in human_df.index:
        minval = int(round(human_df["OFFSET"][row] * SAMPLE_RATE, 0))
        maxval = int(
            round(
                (human_df["OFFSET"][row] +
                 human_df["DURATION"][row]) *
                SAMPLE_RATE,
                0))
        human_arr[minval:maxval] = 1

    human_arr_flipped = 1 - human_arr
    bot_arr_flipped = 1 - bot_arr
    
    true_positive_arr = human_arr * bot_arr
    false_negative_arr = human_arr * bot_arr_flipped
    false_positive_arr = human_arr_flipped * bot_arr
    true_negative_arr = human_arr_flipped * bot_arr_flipped
    IoU_arr = human_arr + bot_arr
    IoU_arr[IoU_arr == 2] = 1

    true_positive_count = np.count_nonzero(
        true_positive_arr == 1) / SAMPLE_RATE
    false_negative_count = np.count_nonzero(
        false_negative_arr == 1) / SAMPLE_RATE
    false_positive_count = np.count_nonzero(
        false_positive_arr == 1) / SAMPLE_RATE
    true_negative_count = np.count_nonzero(
        true_negative_arr == 1) / SAMPLE_RATE
    union_count = np.count_nonzero(IoU_arr == 1) / SAMPLE_RATE

    # Calculating useful values related to tp,fn,fp,tn values

    # Precision = TP/(TP+FP)
    try:
        precision = true_positive_count / \
            (true_positive_count + false_positive_count)

    # Recall = TP/(TP+FN)
        recall = true_positive_count / \
            (true_positive_count + false_negative_count)

    # F1 = 2*(Recall*Precision)/(Recall + Precision)

        f1 = 2 * (recall * precision) / (recall + precision)
        IoU = true_positive_count / union_count
    except BaseException:
        print('''Error calculating statistics, likely due
        to zero division, setting values to zero''')
        f1 = 0
        precision = 0
        recall = 0
        IoU = 0

    # Creating a Dictionary which will be turned into a Pandas Dataframe
    entry = {'FOLDER': folder_name,
             'IN FILE': clip_name,
             'MANUAL ID': clip_class,
             'TRUE POSITIVE': true_positive_count,
             'FALSE POSITIVE': false_positive_count,
             'FALSE NEGATIVE': false_negative_count,
             'TRUE NEGATIVE': true_negative_count,
             'UNION': union_count,
             'PRECISION': precision,
             'RECALL': recall,
             "F1": f1,
             'Global IoU': IoU}

    return pd.DataFrame(entry, index=[0])


# Will have to adjust the isolate function so that it adds a sampling rate
# onto the dataframes.
def automated_labeling_statistics(
        automated_df,
        manual_df,
        stats_type="IoU",
        threshold=0.5):
    """
    Function that will allow users to easily pass in two dataframes of manual
    labels and automated labels, and a dataframe is returned with statistics
    examining the efficacy of the automated labelling system compared to the
    human labels for multiple clips.

    Args:
        automated_df (Dataframe)
            - Dataframe of automated labels of multiple clips.

        manual_df (Dataframe)
            - Dataframe of human labels of multiple clips.

        stats_type (String)
            - String that determines which type of statistics are of interest
            - options: "IoU", "general" 
                - "IoU"
                    - Compares the intersection over union of automated annotations
                      with respect to manual annotations for individual clips.
                - "general"
                    - Consolidates all automated annotations and compares them to all of
                      the manual annotations that have been consolidated across a clip.
            - default: "IoU"

        threshold (Float)
            - Defines a threshold for certain types of statistics such as an
            IoU threshold for determining true positives, false positives, and
            false negatives.
            - default: 0.5

    Returns:
        Dataframe of statistics comparing automated labels and human labels for
        multiple clips.
    """
    # Getting a list of clips
    clips = automated_df["IN FILE"].to_list()
    # Removing duplicates
    clips = list(dict.fromkeys(clips))
    # Initializing the returned dataframe
    statistics_df = pd.DataFrame()

    num_errors = 0
    num_processed = 0

    start_time = time.time()
    # Looping through each audio clip
    for clip in clips:
        num_processed += 1
        clip_automated_df = automated_df[automated_df["IN FILE"] == clip]
        # In case the extension for manual_df is different from the clip extension, just check the name before the extension
        clip_manual_df = manual_df[manual_df["IN FILE"].str.startswith(".".join(clip.split(".")[:-1]))]
        try:
            if stats_type == "general":
                clip_stats_df = clip_general(
                    clip_automated_df, clip_manual_df)
                if statistics_df.empty:
                    statistics_df = clip_stats_df
                else:
                    statistics_df = statistics_df.append(clip_stats_df)
            elif stats_type == "IoU":
                IoU_Matrix = clip_IoU(clip_automated_df, clip_manual_df)
                clip_stats_df = matrix_IoU_Scores(
                    IoU_Matrix, clip_manual_df, threshold)
                if statistics_df.empty:
                    statistics_df = clip_stats_df
                else:
                    statistics_df = statistics_df.append(clip_stats_df)
        except BaseException as e:
            num_errors += 1
            #print("Something went wrong with: " + clip)
            #print(e)
            continue
        if num_processed % 50 == 0:
            print("Processed", num_processed, "clips in", int((time.time() - start_time) * 10) / 10.0, 'seconds')
            start_time = time.time()
    if num_errors > 0:
        print("Something went wrong with", num_errors, "clips out of", len(clips), "clips")
    statistics_df.reset_index(inplace=True, drop=True)
    return statistics_df


def global_dataset_statistics(statistics_df, manual_id = "bird"):
    """
    Function that takes in a dataframe of efficacy statistics for multiple
    clips and outputs their global values.

    Args:
        statistics_df (Dataframe)
            - Dataframe of statistics value for multiple audio clips as
              returned by the function automated_labelling_statistics.
        manual_id (String)
            - String to control the "MANUAL ID" column of the csv file
              format that is used in PyHa. Defaulted to "bird" since the
              package started out with binary bird classification.

    Returns:
        Dataframe of global statistics for the multiple audio clips' labelling.
    """
    tp_sum = statistics_df["TRUE POSITIVE"].sum()
    fp_sum = statistics_df["FALSE POSITIVE"].sum()
    fn_sum = statistics_df["FALSE NEGATIVE"].sum()
    tn_sum = statistics_df["TRUE NEGATIVE"].sum()
    union_sum = statistics_df["UNION"].sum()
    precision = tp_sum / (tp_sum + fp_sum)
    recall = tp_sum / (tp_sum + fn_sum)
    f1 = 2 * (precision * recall) / (precision + recall)
    IoU = tp_sum / union_sum
    entry = {'MANUAL ID': manual_id,
             'PRECISION': round(precision, 6),
             'RECALL': round(recall, 6),
             'F1': round(f1, 6),
             'Global IoU': round(IoU, 6)}
    return pd.DataFrame.from_dict([entry])

def clip_IoU(automated_df, manual_df):
    """
    Function that takes in the manual and automated labels for a clip and
    outputs IoU metrics of each human label with respect to each
    automated label.

    Args:
        automated_df (Dataframe)
            - Dataframe of automated labels for an audio clip.

        manual_df (Dataframe)
             - Dataframe of human labels for an audio clip.

    Returns:
        IoU_Matrix (arr)
            - (human label count) x (automated label count) matrix where each
              row contains the IoU of each automated annotation with respect to
              a human label.
    """

    automated_df.reset_index(inplace=True, drop=True)
    manual_df.reset_index(inplace=True, drop=True)
    # Determining the number of rows in the output numpy array
    manual_row_count = manual_df.shape[0]
    # Determining the number of columns in the output numpy array
    automated_row_count = automated_df.shape[0]

    # Determining the length of the input clip
    duration = automated_df["CLIP LENGTH"].to_list()[0]
    # Determining the sample rate of the input clip
    SAMPLE_RATE = automated_df["SAMPLE RATE"].to_list()[0]

    # Initializing the output array that will contain the clip-by-clip
    # Intersection over Union percentages.
    IoU_Matrix = np.zeros((manual_row_count, automated_row_count))
    # print(IoU_Matrix.shape)

    # Initializing arrays that will represent each of the human and automated
    # labels
    bot_arr = np.zeros((automated_row_count, int(duration * SAMPLE_RATE)))
    human_arr = np.zeros((manual_row_count, int(duration * SAMPLE_RATE)))
    
    # Fill array with automated labels 
    for row in automated_df.index:
        # Determine the beginning of an automated label
        minval = int(round(automated_df["OFFSET"][row] * SAMPLE_RATE, 0))
        # Determining the ending of an automated label
        maxval = int(round((automated_df["OFFSET"][row] + automated_df["DURATION"][row]) * 
                SAMPLE_RATE,0))
        # Placing the label relative to the clip
        bot_arr[row][minval:maxval] = 1

    # Fill array with human labels
    for row in manual_df.index:
        # Determining the beginning of a human label
        minval = int(round(manual_df["OFFSET"][row] * SAMPLE_RATE, 0))
        # Determining the end of a human label
        maxval = int(round((manual_df["OFFSET"][row] + manual_df["DURATION"][row]) *
                SAMPLE_RATE,0))
        # Placing the label relative to the clip
        human_arr[row][minval:maxval] = 1

    # Multiply every row in human by every row in bot
    IoU_Matrix = np.matmul(human_arr, bot_arr.transpose())
    
    # Compare each human annotation to every automated annotation
    for i in range(manual_row_count):
        for j in range(automated_row_count):
            # Skip comparision if there is no intersection, since IoU = 0 anyway 
            if IoU_Matrix[i][j] == 0:
                continue
            # Sum logicial ORR of time bins shared by the human and bot annotations
            IoU_Matrix[i][j] /= np.sum(np.logical_or(human_arr[i], bot_arr[j]))
            IoU_Matrix[i][j] = round(IoU_Matrix[i][j], 4)
    return np.nan_to_num(IoU_Matrix)

def matrix_IoU_Scores(IoU_Matrix, manual_df, threshold):
    """
    Function that takes in the IoU Matrix from the clip_IoU function and outputs
    the number of true positives and false positives, as well as calculating
    the precision, recall, and f1 metrics.

    Args:
        IoU_Matrix (arr)
            - (human label count) x (automated label count) matrix where each
               row contains the IoU of each automated annotation with respect
               to a human label.

        manual_df (Dataframe)
            - Dataframe of human labels for an audio clip.

        threshold (float)
            - IoU threshold for determining true positives, false
              positives, and false negatives.
            - default: 0.5

    Returns:
        Dataframe of clip statistics such as True Positive, False Negative,
        False Positive, Precision, Recall, and F1 values for an audio clip.
    """
    clip_class = manual_df["MANUAL ID"][0]
    audio_dir = manual_df["FOLDER"][0]
    filename = manual_df["IN FILE"][0]
    # TODO make sure that all of these calculations are correct. It is
    # confusing to me that the Precision and Recall scores have a positive
    # correlation. Determining which automated label has the highest IoU across
    # each human label
    automated_label_best_fits = np.max(IoU_Matrix, axis=1)
    # human_label_count = automated_label_best_fits.shape[0]
    # Calculating the number of true positives based off of the passed in
    # thresholds.
    tp_count = automated_label_best_fits[automated_label_best_fits >=
                                         threshold].shape[0]
    # Calculating the number of false negatives from the number of human
    # labels and true positives
    fn_count = automated_label_best_fits[automated_label_best_fits <
                                         threshold].shape[0]

    # Calculating the false positives
    max_val_per_column = np.max(IoU_Matrix, axis=0)
    fp_count = max_val_per_column[max_val_per_column < threshold].shape[0]

    # Calculating the necessary statistics
    try:
        recall = round(tp_count / (tp_count + fn_count), 4)
        precision = round(tp_count / (tp_count + fp_count), 4)
        f1 = round(2 * (recall * precision) / (recall + precision), 4)
    except ZeroDivisionError:
#        print("Division by zero setting precision, recall, and f1 to zero on", filename)
        recall = 0
        precision = 0
        f1 = 0

    entry = {'FOLDER': audio_dir,
             'IN FILE': filename,
             'MANUAL ID': clip_class,
             'TRUE POSITIVE': tp_count,
             'FALSE NEGATIVE': fn_count,
             'FALSE POSITIVE': fp_count,
             'PRECISION': precision,
             'RECALL': recall,
             'F1': f1}

    return pd.DataFrame.from_dict([entry])


def clip_catch(automated_df, manual_df):
    """
    Function that determines whether or not a human label has been found across
    all of the automated labels.

    Args:
        automated_df (Dataframe)
            - Dataframe of automated labels for an audio clip.
        manual_df (Dataframe)
            - Dataframe of human labels for an audio clip.

    Returns:
        Numpy Array of statistics regarding the amount of overlap between the
        manual and automated labels relative to the number of samples.
    """
    # resetting the indices to make this function work
    automated_df.reset_index(inplace=True, drop=True)
    manual_df.reset_index(inplace=True, drop=True)
    # figuring out how many automated labels and human labels exist
    manual_row_count = manual_df.shape[0]
    automated_row_count = automated_df.shape[0]
    # finding the length of the clip as well as the sampling frequency.
    duration = automated_df["CLIP LENGTH"].to_list()[0]
    SAMPLE_RATE = automated_df["SAMPLE RATE"].to_list()[0]
    # initializing the output array, as well as the two arrays used to
    # calculate catch scores
    catch_matrix = np.zeros(manual_row_count)
    bot_arr = np.zeros((int(duration * SAMPLE_RATE)))
    human_arr = np.zeros((int(duration * SAMPLE_RATE)))

    # Determining the automated labelled regions with respect to samples
    # Looping through each human label
    for row in automated_df.index:
        # converting each label into a "pulse" on an array that represents the
        # labels as 0's and 1's on bot array.
        minval = int(round(automated_df["OFFSET"][row] * SAMPLE_RATE, 0))
        maxval = int(
            round(
                (automated_df["OFFSET"][row] +
                 automated_df["DURATION"][row]) *
                SAMPLE_RATE,
                0))
        bot_arr[minval:maxval] = 1

    # Looping through each human label and computing catch =
    # (#intersections)/(#samples in label)
    for row in manual_df.index:
        # Determining the beginning of a human label
        minval = int(round(manual_df["OFFSET"][row] * SAMPLE_RATE, 0))
        # Determining the end of a human label
        maxval = int(
            round(
                (manual_df["OFFSET"][row] +
                 manual_df["DURATION"][row]) *
                SAMPLE_RATE,
                0))
        # Placing the label relative to the clip
        human_arr[minval:maxval] = 1
        # Determining the length of a label with respect to samples
        samples_in_label = maxval - minval
        # Finding where the human label and all of the annotated labels overlap
        intersection = human_arr * bot_arr
        # Determining how many samples overlap.
        intersection_count = np.count_nonzero(intersection == 1)
        # Intersection/length of label
        catch_matrix[row] = round(intersection_count / samples_in_label, 4)
        # resetting the human label
        human_arr[human_arr == 1] = 0

    return catch_matrix


# def dataset_IoU(automated_df,manual_df):
#    """
#    Function that takes in two Pandas dataframes that represent human labels
#    and automated labels.
#    It then runs the clip_IoU function across each clip and appends the best
#    fit IoU score to each labels on the manual dataframe as its output.
#
#    Args:
#        automated_df (Dataframe) - Dataframe of automated labels for multiple
#                                   audio clips.
#        manual_df (Dataframe) - Dataframe of human labels for multiple audio
#                                clips.
#
#    Returns:
#        Dataframe of manual labels with the best fit IoU score as a column.
#    """
#    # Getting a list of clips
#    clips = automated_df["IN FILE"].to_list()
#    # Removing duplicates
#    clips = list(dict.fromkeys(clips))
#    # Initializing the ouput dataframe
#    manual_df_with_IoU = pd.DataFrame()
#    for clip in clips:
#        print(clip)
#        # Isolating a clip from the human and automated dataframes
#        clip_automated_df = automated_df[automated_df["IN FILE"] == clip]
#        clip_manual_df = manual_df[manual_df["IN FILE"] == clip]
#        # Calculating the IoU scores of each human label.
#        IoU_Matrix = clip_IoU(clip_automated_df,clip_manual_df)
#        # Finding the best automated IoU score with respect to each label
#        automated_label_best_fits = np.max(IoU_Matrix,axis=1)
#        clip_manual_df["IoU"] = automated_label_best_fits
#        # Appending on the best fit IoU score to each human label
#        if manual_df_with_IoU.empty == True:
#            manual_df_with_IoU = clip_manual_df
#        else:
#            manual_df_with_IoU = manual_df_with_IoU.append(clip_manual_df)
#    # Adjusting the indices.
#    manual_df_with_IoU.reset_index(inplace = True, drop = True)
#    return manual_df_with_IoU


# def class_IoU_Statistics(automated_df,manual_df,threshold = 0.5):
#    """
#    Wrapper function that takes matrix_IoU_Scores across multiple clips from a
#    class. Allows user to modify the threshold that determines whether or not
#    a label is a true positive.

#    Args:
#         automated_df (Dataframe)
#             - Dataframe of automated labels for multiple
#                                   audio clips.

#         manual_df (Dataframe)
#             - Dataframe of human labels for multiple audio clips.

#         threshold (float)
#             - IoU threshold for determining true positives, false positives,
#               and false negatives.

#    Returns:
#        Dataframe of IoU statistics for multiple audio clips.
#    """
#    # isolating the names of the clips that have been labelled into an array.
#    clips = automated_df["IN FILE"].to_list()
#    clips = list(dict.fromkeys(clips))
#    # initializing the output Pandas dataframe
#    # Looping through all of the clips
#    for clip in clips:
#        print(clip)
#        clip_automated_df = automated_df[automated_df["IN FILE"] == clip]
#        clip_manual_df = manual_df[manual_df["IN FILE"] == clip]
#        # Computing the IoU Matrix across a specific clip
#        IoU_Matrix = clip_IoU(clip_automated_df,clip_manual_df)
#        # Calculating the best fit IoU to each label for the clip
#        clip_stats_df = matrix_IoU_Scores(IoU_Matrix,clip_manual_df,threshold)
#        # adding onto the output array.
#        if IoU_Statistics.empty == True:
#            IoU_Statistics = clip_stats_df
#        else:
#            IoU_Statistics = IoU_Statistics.append(clip_stats_df)
#    IoU_Statistics.reset_index(inplace = True, drop = True)
#    return IoU_Statistics

# Consider adding in a new manual_id parameter here
def global_statistics(statistics_df, manual_id = 'N/A'):
    """
    Function that takes the output of dataset_IoU Statistics and outputs a
    global count of true positives and false positives, as well as computing \
    the precision, recall, and f1 metrics across the dataset.

    Args:
        statistics_df (Dataframe)
            - Dataframe of matrix IoU scores for multiple clips.
        
        manual_id (String)
            - String to control the "MANUAL ID" column of the csv file
              format that is used in PyHa.
            - default: "N/A"

    Returns:
        Dataframe of global IoU statistics which include the number of true
        positives, false positives, and false negatives. Contains Precision,
        Recall, and F1 metrics as well
    """

    #data_class = statistics_df["MANUAL ID"][0]
    # taking the sum of the number of true positives and false positives.
    tp_sum = statistics_df["TRUE POSITIVE"].sum()
    fn_sum = statistics_df["FALSE NEGATIVE"].sum()
    fp_sum = statistics_df["FALSE POSITIVE"].sum()
    # calculating the precision, recall, and f1
    try:
        precision = tp_sum / (tp_sum + fp_sum)
        recall = tp_sum / (tp_sum + fn_sum)
        f1 = 2 * (precision * recall) / (precision + recall)
    except ZeroDivisionError:
        print('''Error in calculating Precision, Recall, and F1. Likely due to
        zero division, setting values to zero''')
        precision = 0
        recall = 0
        f1 = 0
    # building a dictionary of the above calculations
    entry = {'MANUAL ID': manual_id,
             'TRUE POSITIVE': tp_sum,
             'FALSE NEGATIVE': fn_sum,
             'FALSE POSITIVE': fp_sum,
             'PRECISION': round(precision, 4),
             'RECALL': round(recall, 4),
             'F1': round(f1, 4)}
    # returning the dictionary as a pandas dataframe
    return pd.DataFrame.from_dict([entry])


def dataset_Catch(automated_df, manual_df):
    """
    Function that determines the overlap of each human label with respect to
    all of the human labels in a clip across a large number of clips.

    Args:
        automated_df (Dataframe)
            - Dataframe of automated labels for multiple audio clips.

        manual_df (Dataframe)
            - Dataframe of human labels for multiple audio clips.

    Returns:
        Dataframe of human labels with a column for the catch values of each
        label.
    """
    # Getting a list of clips
    clips = automated_df["IN FILE"].to_list()
    # Removing duplicates
    clips = list(dict.fromkeys(clips))
    # Initializing the ouput dataframe
    manual_df_with_Catch = pd.DataFrame()
    # Looping through all of the audio clips that have been labelled.
    for clip in clips:
        print(clip)
        # Isolating the clips from both the automated and human dataframes
        clip_automated_df = automated_df[automated_df["IN FILE"] == clip]
        clip_manual_df = manual_df[manual_df["IN FILE"] == clip]
        # Calling the function that calculates the catch over a specific clip
        Catch_Array = clip_catch(clip_automated_df, clip_manual_df)
        # Appending the catch values per label onto the manual dataframe
        clip_manual_df["Catch"] = Catch_Array
        if manual_df_with_Catch.empty:
            manual_df_with_Catch = clip_manual_df
        else:
            manual_df_with_Catch = manual_df_with_Catch.append(clip_manual_df)
    # Resetting the indices
    manual_df_with_Catch.reset_index(inplace=True, drop=True)
    return manual_df_with_Catch


def clip_statistics(
        automated_df,
        manual_df, 
        stats_type = "IoU", 
        threshold = 0.5):
    """
    Function to generate a dataframe containing efficacy statistics of automated
    labeling compared to human labels for multiple classes.

    Calls automated_labeling_statistics on individual classes to identify
    the overlapping human and automated labels, then concats the dataframes.

    Args: 
        automated_df (Dataframe)
            - Dataframe of automated labels for multiple classes.

        manual_df (Dataframe)
            - Dataframe of human labels for multiple classes.
        
        stats_type (String)
            - String that determines which type of statistics are of interest
            - options: "IoU", "general" 
                - "IoU"
                    - Compares the intersection over union of automated annotations
                      with respect to manual annotations for individual clips.
                - "general"
                    - Consolidates all automated annotations and compares them to all of
                      the manual annotations that have been consolidated across a clip.
            - default: "IoU"

        threshold (Float)
            - Defines a threshold for certain types of statistics such as an
              IoU threshold for determining true positives, false positives, and
              false negatives.
            - default: 0.5
    
    Returns: 
        Dataframe with clip overlap statistics comparing automated and human 
        labeling for multiple classes.
    """
    # Creating a list of classes from the automated dataframe
    automated_class_list = automated_df["MANUAL ID"].to_list()
    automated_class_list = list(dict.fromkeys(automated_class_list))
    # Creating a list of classes from the manual dataframe
    manual_class_list = manual_df["MANUAL ID"].to_list()
    manual_class_list = list(dict.fromkeys(manual_class_list))
    # Finding the intersection between the manual and automated classes
    class_list = np.intersect1d(automated_class_list,manual_class_list)
    
    # Initializing the output dataframe
    clip_statistics = pd.DataFrame()
    # Looping through each class and comparing the automated labels to the manual labels
    for class_ in class_list:
        #print(class_)
        # isolating the current class of interest
        temp_manual_class_df = manual_df[manual_df["MANUAL ID"] == class_]
        temp_automated_class_df = automated_df[automated_df["MANUAL ID"] == class_]
        # The case if clip_statistics hasn't been filled yet
        if clip_statistics.empty:
            clip_statistics = automated_labeling_statistics(temp_automated_class_df, temp_manual_class_df, stats_type = stats_type, threshold = threshold)
        else:
            temp_df = automated_labeling_statistics(temp_automated_class_df, temp_manual_class_df, stats_type = stats_type, threshold = threshold)
            clip_statistics = clip_statistics.append(temp_df)
    clip_statistics.reset_index(inplace=True,drop=True)
    return clip_statistics

def class_statistics(clip_statistics):
    """
    Function that takes in a dataframe of efficacy statistics for multiple 
    classes and outputs global efficacy values for each class.

    Calls global_statistics on individual classes, then concats the dataframes.

    Args: 
        clip_statistics (Dataframe)
            - Dataframe of multi-class statistics values for audio clips as 
              returned by the function clip_statistics.
    
    Returns:
        Dataframe of global efficacy statistics for multiple classes.
    """
    # Initializing the output dataframe
    class_statistics = pd.DataFrame()
    # creating a list of the unique classes being passed in.
    class_list = clip_statistics["MANUAL ID"].to_list()
    class_list = list(dict.fromkeys(class_list))
    for class_ in class_list:
        #print(class_)
        # isolating the current class of interest
        class_df = clip_statistics[clip_statistics["MANUAL ID"] == class_]
        if class_statistics.empty:
            class_statistics = global_statistics(class_df, manual_id = class_)
        else:
            temp_df = global_statistics(class_df, manual_id = class_)
            class_statistics = class_statistics.append(temp_df)
    class_statistics.reset_index(inplace=True,drop=True)
    return class_statistics



from .birdnet_lite.analyze import analyze
from .microfaune_package.microfaune.detection import RNNDetector
from .microfaune_package.microfaune import audio
from .tweetynet_package.tweetynet.TweetyNetModel import TweetyNetModel
from .tweetynet_package.tweetynet.Load_data_functions import compute_features, predictions_to_kaleidoscope
import os
import torch
import librosa
import pandas as pd
import scipy.signal as scipy_signal
import numpy as np
from math import ceil
from copy import deepcopy

def build_isolation_parameters_microfaune(
        technique,
        threshold_type,
        threshold_const,
        threshold_min=0,
        window_size=1.0,
        chunk_size=2.0):
    """
    Wrapper function for all audio isolation techniques (Steinberg, Simple, 
    Stack, Chunk). Will call the respective function of each technique
    based on isolation_parameters "technique" key.

    Args:
        technique (string)
            - Chooses which of the four isolation techniques to deploy
            - options: "steinberg", "simple", "stack", "chunk"

        threshold_type (string)
            - Chooses how to derive a threshold from local score arrays
            - options: "mean", "median", "standard deviation", "pure"

        threshold_const (float)
            - Multiplier for "mean", "median", and "standard deviation". Acts
              as threshold for "pure"

        threshold_min (float)
            - Serves as a minimum barrier of entry for a local score to be
              considered a positive ID of a class.
            - default: 0

        window_size (float)
            - determines how many seconds around a positive ID local score
              to build an annotation.

        chunk_size (float)
            - determines the length of annotation when using "chunk"
              isolation technique

    Returns:
        isolation_parameters (dict)
            - Python dictionary that controls how to go about isolating
              automated labels from audio.
    """
    isolation_parameters = {
        "technique": technique,
        "threshold_type": threshold_type,
        "threshold_const": threshold_const,
        "threshold_min": threshold_min,
        "chunk_size": chunk_size
    }

    if window_size != 1.0 and technique != "steinberg":
        print('''Warning: window_size is dedicated to the steinberg isolation
        technique. Won't affect current technique.''')
    if chunk_size != 2.0 and technique != "chunk":
        print('''Warning: chunk_size is dedicated to the chunk technique.
        Won't affect current technique.''')

    return isolation_parameters


def isolate(
        local_scores,
        SIGNAL,
        SAMPLE_RATE,
        audio_dir,
        filename,
        isolation_parameters,
        manual_id="bird",
        normalize_local_scores=False):
    """
    Wrapper function for all of Microfaune's audio isolation techniques 
    (Steinberg, Simple, Stack, Chunk). Will call the respective function of each technique based on 
    isolation_parameters "technique" key.

    Args:
        local_scores (list of floats)
            - Local scores of the audio clip as determined by
              Microfaune Recurrent Neural Network.

        SIGNAL (list of ints)
            - Samples that make up the audio signal.

        SAMPLE_RATE (int)
            - Sampling rate of the audio clip, usually 44100.

        audio_dir (string)
            - Directory of the audio clip.

        filename (string)
            - Name of the audio clip file.

        isolation_parameters (dict)
            - Python Dictionary that controls the various label creation
              techniques.

    Returns:
        Dataframe of automated labels for the audio clip based on passed in
        isolation technique.
    """

    # normalize the local scores so that the max value is 1.
    #if normalize_local_scores:
    #    local_scores_max = max(local_scores)
    #    for ndx in range(len(local_scores)):
    #        local_scores[ndx] = local_scores[ndx] / local_scores_max
    # initializing the output dataframe that will contain labels across a
    # single clip
    isolation_df = pd.DataFrame()

    # deciding which isolation technique to deploy for a given clip based on
    # the technique isolation parameter
    if isolation_parameters["technique"] == "simple":
        isolation_df = simple_isolate(
            local_scores,
            SIGNAL,
            SAMPLE_RATE,
            audio_dir,
            filename,
            isolation_parameters,
            manual_id=manual_id)
    elif isolation_parameters["technique"] == "steinberg":
        isolation_df = steinberg_isolate(
            local_scores,
            SIGNAL,
            SAMPLE_RATE,
            audio_dir,
            filename,
            isolation_parameters,
            manual_id=manual_id)
    elif isolation_parameters["technique"] == "stack":
        isolation_df = stack_isolate(
            local_scores,
            SIGNAL,
            SAMPLE_RATE,
            audio_dir,
            filename,
            isolation_parameters,
            manual_id=manual_id)
    elif isolation_parameters["technique"] == "chunk":
        isolation_df = chunk_isolate(
            local_scores,
            SIGNAL,
            SAMPLE_RATE,
            audio_dir,
            filename,
            isolation_parameters,
            manual_id=manual_id)

    return isolation_df


def threshold(local_scores, isolation_parameters):
    """
    Takes in the local score array output from a neural network and determines
    the threshold at which we determine a local score to be a positive
    ID of a class of interest. Most proof of concept work is dedicated to bird
    presence. Threshold is determined by "threshold_type" and "threshold_const"
    from the isolation_parameters dictionary.

    Args:
        local_scores (list of floats)
            - Local scores of the audio clip as determined by Microfaune
              Recurrent Neural Network.

        isolation_parameters (dict)
            - Python Dictionary that controls the various label creation
              techniques.

    Returns:
        thresh (float)
            - threshold at which the local scores in the local score array of
              an audio clip will be viewed as a positive ID.
    """
    if isolation_parameters["threshold_type"] == "median":
        thresh = np.median(local_scores) \
            * isolation_parameters["threshold_const"]
    elif (isolation_parameters["threshold_type"] == "mean" or
            isolation_parameters["threshold_type"] == "average"):
        thresh = np.mean(local_scores) \
            * isolation_parameters["threshold_const"]
    elif isolation_parameters["threshold_type"] == "standard deviation":
        thresh = np.mean(local_scores) + \
            (np.std(local_scores) * isolation_parameters["threshold_const"])
    elif isolation_parameters["threshold_type"] == "pure":
        thresh = isolation_parameters["threshold_const"]
        if thresh < 0:
            print("Threshold is less than zero, setting to zero")
            thresh = 0
        elif thresh > 1:
            print("Threshold is greater than one, setting to one.")
            thresh = 1
    return thresh


def steinberg_isolate(
        local_scores,
        SIGNAL,
        SAMPLE_RATE,
        audio_dir,
        filename,
        isolation_parameters,
        manual_id="bird"):
    """
    Technique developed by Gabriel Steinberg that attempts to take the local 
    score array output of a neural network and lump local scores together in
    a way to produce automated labels based on a class across an audio clip.

    Technique Pseudocode:

    Loop through local score array:
        if current local score > (threshold and threshold_min):
            build an annotation with current local score at the center with
            +- window_size/2 seconds around current local score.
        else:
            continue
    extra logic handles overlap if a local score meets the criteria within
    the "window_size" from a prior local score

    Args:
        local_scores (list of floats)
            - Local scores of the audio clip as determined by RNNDetector.

        SIGNAL (list of ints)
            - Samples from the audio clip.

        SAMPLE_RATE (int)
            - Sampling rate of the audio clip, usually 44100.

        audio_dir (string)
            - Directory of the audio clip.

        filename (string)
            - Name of the audio clip file.

        isolation_parameters (dict)
            - Python Dictionary that controls the various label creation
              techniques.

        manual_id (string)
            - controls the name of the class written to the pandas dataframe

    Returns:
        Pandas Dataframe of automated labels for the audio clip.
    """
    # calculate original duration
    old_duration = len(SIGNAL) / SAMPLE_RATE
    # create entry for audio clip
    entry = {'FOLDER': audio_dir,
             'IN FILE': filename,
             'CHANNEL': 0,
             'CLIP LENGTH': old_duration,
             'SAMPLE RATE': SAMPLE_RATE,
             'OFFSET': [],
             'MANUAL ID': []}

    # calculating threshold that will define how labels are created in current
    # audio clip
    thresh = threshold(local_scores, isolation_parameters)
    # how many samples one local score represents
    samples_per_score = len(SIGNAL) // len(local_scores)
    
    # Calculating local scores that are at or above threshold
    thresh_scores = local_scores >= max(thresh, isolation_parameters["threshold_min"])
    
    # if statement to check if window size is smaller than time between two local scores
    # (as a safeguard against problems that can occur)
    if (int(isolation_parameters["window_size"] / 2 * SAMPLE_RATE) * 2 >= samples_per_score):
        # Set up to find the starts and ends of clips (not considering window)
        thresh_scores = np.append(thresh_scores, [0])
        rolled_scores = np.roll(thresh_scores, 1)
        rolled_scores[0] = 0
        diff_scores = thresh_scores - rolled_scores

        # Logic for finding the starts and ends:
        # If thresh_scores = [1 1 1 1 0 0 0 1 1], then
        # thresh_scores becomes [1 1 1 1 0 0 0 1 1 0] and
        # rolled_scores are [0 1 1 1 1 0 0 0 1 1]. Subtracting
        # yields [1 0 0 0 -1 0 0 1 0 -1]. The 1s are the starts of the clips,
        # and the -1s are 1 past the ends of the clips
        
        # Adds the "window" to each annotation
        starts = np.where(diff_scores == 1)[0] * samples_per_score - int(isolation_parameters["window_size"] / 2 * SAMPLE_RATE)
        ends = np.where(diff_scores == -1)[0] - 1
        ends = ends * samples_per_score + int(isolation_parameters["window_size"] / 2 * SAMPLE_RATE)
        
        # Does not continue if no annotations exist
        if (len(starts) == 0):
            return pd.DataFrame.from_dict({'FOLDER': [],       \
                                           'IN FILE': [],      \
                                           'CHANNEL': [],      \
                                           'CLIP LENGTH': [],  \
                                           'SAMPLE RATE': [],  \
                                           'OFFSET': [],       \
                                           'MANUAL ID': []})
        
        # Checks annotations for any overlap, and removes if so
        i = 0
        while True:
            if (i == len(ends) - 1):
                break
            if (starts[i + 1] < ends[i]):
                ends = np.delete(ends, i)
                starts = np.delete(starts, i + 1)
            else:
                i += 1
        
        # Correcting bounds
        starts[0] = max(0, starts[0])
        ends[-1] = min(len(SIGNAL), ends[-1])
        
        # Calculates offsets and durations from starts and ends
        entry['OFFSET'] = starts * 1.0 / SAMPLE_RATE
        entry['DURATION'] = ends - starts
        entry['DURATION'] = entry['DURATION'] * 1.0 / SAMPLE_RATE
        
        # Assigns manual ids to all annotations
        entry['MANUAL ID'] = np.full(entry['OFFSET'].shape, manual_id)
    else:
        # Simply assigns each 1 in thresh scores to be its own window if windows are too small
        entry['OFFSET'] = np.where(thresh_scores == 1)[0] * samples_per_score / SAMPLE_RATE - isolation_parameters["window_size"] / 2
        entry['DURATION'] = np.full(entry['OFFSET'].shape, isolation_parameters["window_size"] * 1.0)
        if (entry['OFFSET'] < 0):
            entry['OFFSET'][0] = 0
            entry['DURATION'][0] = isolation_parameters["window_size"] * 0.5
        entry['MANUAL ID'] = np.full(entry['OFFSET'].shape, manual_id)
    
    # returning pandas dataframe from dictionary constructed with all of the
    # annotations
    return pd.DataFrame.from_dict(entry)

def simple_isolate(
        local_scores,
        SIGNAL,
        SAMPLE_RATE,
        audio_dir,
        filename,
        isolation_parameters,
        manual_id="bird"):
    """
    Technique suggested by Irina Tolkova, implemented by Jacob Ayers. 
    Attempts to produce automated annotations of an audio clip based
    on local score array outputs from a neural network.

    Technique Pseudocode:

    Loop through local score array:
        if current local score > (threshold and threshold_min)
        and annotation start = 0:
            start annotation
        else if current local score < thresh and annotation start = 1:
            end annotation
        else:
            continue

    Args:
        local_scores (list of floats)
            - Local scores of the audio clip as determined by RNNDetector.

        SIGNAL (list of ints)
            - Samples from the audio clip.

        SAMPLE_RATE (int)
            - Sampling rate of the audio clip, usually 44100.

        audio_dir (string)
            - Directory of the audio clip.

        filename (string)
            - Name of the audio clip file.

        isolation_parameters (dict)
            - Python Dictionary that controls the various label creation
              techniques.

        manual_id (string)
            - controls the name of the class written to the pandas dataframe

    Returns:
        Pandas Dataframe of automated labels for the audio clip.
    """
    # Calculating threshold that defines the creation of the automated labels
    # for an audio clip
    thresh = threshold(local_scores, isolation_parameters)

    # calculate original duration
    old_duration = len(SIGNAL) / SAMPLE_RATE

    entry = {'FOLDER': audio_dir,
             'IN FILE': filename,
             'CHANNEL': 0,
             'CLIP LENGTH': old_duration,
             'SAMPLE RATE': SAMPLE_RATE,
             'OFFSET': [],
             'DURATION': [],
             'MANUAL ID': []}

    # how many samples one score represents
    # Scores meaning local scores
    samples_per_score = len(SIGNAL) // len(local_scores)
    # local_score * samples_per_score / sample_rate
    time_per_score = samples_per_score / SAMPLE_RATE

    # Calculating local scores that are at or above threshold
    thresh_scores = local_scores >= max(thresh, isolation_parameters["threshold_min"])
    
    # Set up to find the starts and ends of clips
    thresh_scores = np.append(thresh_scores, [0])
    rolled_scores = np.roll(thresh_scores, 1)
    rolled_scores[0] = 0
    
    # Logic for finding starts and ends given in steinberg isolate
    diff_scores = thresh_scores - rolled_scores
    
    # Calculates offsets and durations from difference
    entry['OFFSET'] = np.where(diff_scores == 1)[0] * time_per_score * 1.0
    entry['DURATION'] = np.where(diff_scores == -1)[0] * time_per_score - entry['OFFSET']
    
    # Assigns manual ids to all annotations
    entry['MANUAL ID'] = np.full(entry['OFFSET'].shape, manual_id)
    
    # returning pandas dataframe from dictionary constructed with all of the
    # annotations
    return pd.DataFrame.from_dict(entry)

def stack_isolate(
        local_scores,
        SIGNAL,
        SAMPLE_RATE,
        audio_dir,
        filename,
        isolation_parameters,
        manual_id="bird"):
    """
    Technique created by Jacob Ayers. Attempts to produce automated
    annotations of an audio clip base on local score array outputs
    from a neural network.

    Technique Pseudocode:

    Loop through local score array:
        if current local score > (threshold and threshold_min):
            if annotation start false:
                set annotation start true
            push to stack counter
        else if current local score < thresh and annotation start true:
            pop from stack counter
            if stack counter = 0:
                end annotation
        else:
            continue

    Args:
        local_scores (list of floats)
            - Local scores of the audio clip as determined by RNNDetector.

        SIGNAL (list of ints)
            - Samples from the audio clip.

        SAMPLE_RATE (int)
            - Sampling rate of the audio clip, usually 44100.

        audio_dir (string)
            - Directory of the audio clip.

        filename (string)
            - Name of the audio clip file.

        isolation_parameters (dict)
            - Python Dictionary that controls the various label creation
              techniques.

        manual_id (string)
            - controls the name of the class written to the pandas dataframe

    Returns:
        Pandas Dataframe of automated labels for the audio clip.
    """
    # configuring the threshold based on isolation parameters
    thresh = threshold(local_scores, isolation_parameters)

    # calculate original duration
    old_duration = len(SIGNAL) / SAMPLE_RATE

    # initializing a dictionary that will be used to construct the output
    # pandas dataframe.
    entry = {'FOLDER': audio_dir,
             'IN FILE': filename,
             'CHANNEL': 0,
             'CLIP LENGTH': old_duration,
             'SAMPLE RATE': SAMPLE_RATE,
             'OFFSET': [],
             'DURATION': [],
             'MANUAL ID': []}

    # how many samples one score represents
    # Scores meaning local scores
    samples_per_score = len(SIGNAL) // len(local_scores)
    # local_score * samples_per_score / sample_rate
    # constant that will be used to convert from local score indices to
    # annotation start/stop values.
    time_per_score = samples_per_score / SAMPLE_RATE

    # Calculating local scores that are at or above threshold
    thresh_scores = local_scores >= max(thresh, isolation_parameters["threshold_min"])
    
    # Set up to find the starts and ends of clips
    thresh_scores = np.append(thresh_scores, [0])
    rolled_scores = np.roll(thresh_scores, 1)
    rolled_scores[0] = 0

    # Logic for finding starts and ends given in steinberg isolate
    diff_scores = thresh_scores - rolled_scores

    starts = np.where(diff_scores == 1)[0]
    ends = np.where(diff_scores == -1)[0]
    
    # Stack algorithm: considers a stack counter, and
    # updates stack counter between annotations (+1 for every 
    # entry above the threshold, -1 for below); Combines annotations
    # in this way, along with any adjacent annotations (where stack
    # counter is 0 for one value between annotations).
    i = 0
    while i < len(ends):
        stack_counter = ends[i] - starts[i]
        new_end = ends[i] + stack_counter
        while (i < len(ends) - 1 and starts[i + 1] <= new_end):
            stack_counter -= starts[i + 1] - ends[i]
            stack_counter += ends[i + 1] - starts[i + 1]
            ends = np.delete(ends, i)
            starts = np.delete(starts, i + 1)
            new_end = ends[i] + stack_counter
        ends[i] = new_end
        i += 1
    
    # Addressing situation where end goes above max length of local scores
    ends[-1] = min(len(local_scores) - 1, ends[-1])

    # Deletes annotation if it starts on the
    # last local score
    if (starts[-1] == len(local_scores) - 1):
        starts = np.delete(starts, len(starts) - 1)
        ends = np.delete(ends, len(ends) - 1)
    
    # Calculates offsets and durations from starts/ends
    entry['OFFSET'] = starts * time_per_score
    entry['DURATION'] = ends - starts
    entry['DURATION'] = entry['DURATION'] * time_per_score
    
    # Assigns manual ids to all annotations
    entry['MANUAL ID'] = np.full(entry['OFFSET'].shape, manual_id)
    
    # returning pandas dataframe from dictionary constructed with all of the
    # annotations
    return pd.DataFrame.from_dict(entry)

# TODO
# Make it so that a user has the option of an overlap between the chunks.
# Make it so that a user can choose how many samples have to be above the
# threshold in order to consider a chunk to be good or not.
# Give the option to combine annotations that follow one-another.


def chunk_isolate(
        local_scores,
        SIGNAL,
        SAMPLE_RATE,
        audio_dir,
        filename,
        isolation_parameters,
        manual_id="bird"):
    """
    Technique created by Jacob Ayers. Attempts to produce automated 
    annotations of an audio clip based on local score array outputs
    from a neural network.

    Technique Pseudocode:

    number of chunks = clip length / "chunk_size"
    Loop through number of chunks:
        if max(local score chunk) > (threshold and "threshold_min"):
            set the chunk as an annotation
        else:
            continue

    Args:
        local_scores (list of floats)
            - Local scores of the audio clip as determined by RNNDetector.

        SIGNAL (list of ints)
            - Samples from the audio clip.

        SAMPLE_RATE (int)
            - Sampling rate of the audio clip, usually 44100.

        audio_dir (string)
            - Directory of the audio clip.

        filename (string)
            - Name of the audio clip file.

        isolation_parameters (dict)
            - Python Dictionary that controls the various label creation
              techniques.

        manual_id (string)
            - controls the name of the class written to the pandas dataframe

    Returns:
        Pandas Dataframe of automated labels for the audio clip.
    """
    # configuring the threshold based on isolation parameters
    thresh = threshold(local_scores, isolation_parameters)

    # calculate original duration
    old_duration = len(SIGNAL) / SAMPLE_RATE

    # initializing the dictionary for the output pandas dataframe
    entry = {'FOLDER': audio_dir,
             'IN FILE': filename,
             'CHANNEL': 0,
             'CLIP LENGTH': old_duration,
             'SAMPLE RATE': SAMPLE_RATE,
             'OFFSET': [],
             'DURATION': [],
             'MANUAL ID': manual_id}

    # calculating the number of chunks that define an audio clip
    chunk_count = ceil(
        len(SIGNAL) / (isolation_parameters["chunk_size"] * SAMPLE_RATE))
    # calculating the number of local scores per second
    scores_per_second = len(local_scores) / old_duration
    # calculating the chunk size with respect to the local score array
    local_scores_per_chunk = scores_per_second * \
        isolation_parameters["chunk_size"]
    
    # Creates indices for starts of chunks using np.linspace
    # which creates even splits across a range, and then is
    # treated as int (rounds down)
    chunk_starts_float = np.linspace(start = 0, stop = chunk_count * local_scores_per_chunk, num = chunk_count, endpoint = False)
    chunk_starts = chunk_starts_float.copy().astype(int)
    
    # Deletes the first element of the array (0) to
    # avoid empty array
    chunk_starts = np.delete(chunk_starts, 0)
    
    # Creates chunked scores based on starts
    # Finds max value of each chunked array
    chunked_scores = np.array(list(map(np.amax, np.split(local_scores, chunk_starts))))
    
    # Finds which chunks are above threshold, and creates indices based on that
    thresh_scores = chunked_scores >= max(thresh, isolation_parameters["threshold_min"])
    chunk_indices = np.where(thresh_scores == 1)[0]
    
    # Assigns offset values based on float values of the starts
    entry['OFFSET'] = chunk_starts_float[chunk_indices] / scores_per_second
    
    # Creates durations based on float values of chunk starts
    all_chunk_durs = np.roll(chunk_starts_float, -1) / scores_per_second - chunk_starts_float / scores_per_second
    all_chunk_durs[-1] = len(local_scores) / scores_per_second - chunk_starts_float[-1] / scores_per_second
    entry['DURATION'] = all_chunk_durs[chunk_indices]

    # Assigns manual ids to all annotations
    entry['MANUAL ID'] = np.full(entry['OFFSET'].shape, manual_id)
    
    # returning pandas dataframe from dictionary constructed with all of the
    # annotations
    return pd.DataFrame.from_dict(entry)


def generate_automated_labels_birdnet(audio_dir, isolation_parameters):
    """
    Function that generates the bird labels for an audio file or across a
    folder using the BirdNet-Lite model

    Args:
        audio_dir (string)
            - Directory with wav audio files. Can be an individual file
              as well.

        isolation_parameters (dict)
            - Python Dictionary that controls the various label creation
              techniques. The keys it accepts are :
              - output_path (string)
                - Path to output folder. By default results are written into 
                  the input folder
                - default: None

              - lat (float)
                - Recording location latitude
                - default: -1 (ignore)

              - lon (float)
                - Recording location longitude
                - default: -1 (ignore)

              - week (int)
                - Week of the year when the recording was made
                - Values in [1, 48] (4 weeks per month) 
                - default: -1 (ignore)

              - overlap (float)
                - Overlap in seconds between extracted spectrograms
                - Values in [0.5, 1.5]
                - default: 0.0

              - sensitivity (float)
                - Detection sensitivity. Higher values result in higher sensitivity
                - Values in [0.5, 1.5] 
                - default: 1.0

              - min_conf (float)
                - Minimum confidence threshold
                - Values in [0.01, 0.99]
                - default: 0.1

              - custom_list (string)
                - Path to text file containing a list of species
                - default: '' (not used if not provided)

              - filetype (string)
                - Filetype of soundscape recordings
                - default: 'wav'

              - num_predictions (int)
                - Defines maximum number of written predictions in a given 3s segment
                - default: 10

              - write_to_csv (bool)
                - Set whether or not to write output to CSV
                - default: False

    Returns:
        Dataframe of automated labels for the audio clip(s) in audio_dir.
    """
    annotations = analyze(audio_path=audio_dir, **isolation_parameters)
    return annotations

def generate_automated_labels_microfaune(
        audio_dir,
        isolation_parameters,
        ml_model = "microfaune",
        manual_id="bird",
        weight_path=None,
        normalized_sample_rate=44100,
        normalize_local_scores=False):
    """
    Function that applies isolation technique on the local scores generated
    by the Microfaune mode across a folder of audio clips. It is determined
    by the isolation_parameters dictionary.

    Args:
        audio_dir (string)
            - Directory with wav audio files.

        isolation_parameters (dict)
            - Python Dictionary that controls the various label creation
              techniques.

        manual_id (string)
            - controls the name of the class written to the pandas dataframe

        weight_path (string)
            - File path of weights to be used by the RNNDetector for
              determining presence of bird sounds.

        normalized_sample_rate (int)
            - Sampling rate that the audio files should all be normalized to.

    Returns:
        Dataframe of automated labels for the audio clips in audio_dir.
    """

    # init detector
    # Use Default Microfaune Detector
    # TODO
    # Expand to neural networks beyond just microfaune
    #Add flag to work for creating tweetynet model.
    if weight_path is None:
        detector = RNNDetector()
    # Use Custom weights for Microfaune Detector
    else:
        detector = RNNDetector(weight_path)
        # print("model \"{}\" does not exist".format(ml_model))
        # return None

    # init labels dataframe
    annotations = pd.DataFrame()
    # generate local scores for every bird file in chosen directory
    for audio_file in os.listdir(audio_dir):
        # skip directories
        if os.path.isdir(audio_dir + audio_file):
            continue

        # Reading in the audio files using librosa, converting to single channeled data with original sample rate
        # Reason for the factor for the signal is explained here: https://stackoverflow.com/questions/53462062/pyaudio-bytes-data-to-librosa-floating-point-time-series
        # Librosa scales down to [-1, 1], but the models require the range [-32768, 32767]
        try:
            SIGNAL, SAMPLE_RATE = librosa.load(audio_dir + audio_file, sr=None, mono=True)
            SIGNAL = SIGNAL * 32768
        except BaseException:
            print("Failed to load", audio_file)
            continue

        # downsample the audio if the sample rate isn't 44.1 kHz
        # Force everything into the human hearing range.
        # May consider reworking this function so that it upsamples as well
        try:
            if SAMPLE_RATE != normalized_sample_rate:
                rate_ratio = normalized_sample_rate / SAMPLE_RATE
                SIGNAL = scipy_signal.resample(
                    SIGNAL, int(len(SIGNAL) * rate_ratio))
                SAMPLE_RATE = normalized_sample_rate
        except:
            print("Failed to Downsample" + audio_file)
            # resample produces unreadable float32 array so convert back
            # SIGNAL = np.asarray(SIGNAL, dtype=np.int16)

        # print(SIGNAL.shape)
        # convert stereo to mono if needed
        # Might want to compare to just taking the first set of data.
        if len(SIGNAL.shape) == 2:
            SIGNAL = SIGNAL.sum(axis=1) / 2
        # detection
        try:
            microfaune_features = detector.compute_features([SIGNAL])
            global_score, local_scores = detector.predict(microfaune_features)
        except BaseException as e:
            print("Error in detection, skipping", audio_file)
            print(e)
            continue

        # get duration of clip
        duration = len(SIGNAL) / SAMPLE_RATE

        try:
            # Running moment to moment algorithm and appending to a master
            # dataframe.
            new_entry = isolate(
                local_scores[0],
                SIGNAL,
                SAMPLE_RATE,
                audio_dir,
                audio_file,
                isolation_parameters,
                manual_id=manual_id,
                normalize_local_scores=normalize_local_scores)
            # print(new_entry)
            if annotations.empty:
                annotations = new_entry
            else:
                annotations = annotations.append(new_entry)
        except BaseException as e:
            print("Error in isolating bird calls from", audio_file)
            print(e)
            continue
    # Quick fix to indexing
    annotations.reset_index(inplace=True, drop=True)
    return annotations

def generate_automated_labels_tweetynet(
        audio_dir,
        isolation_parameters,
        manual_id="bird",
        weight_path=None,
        normalized_sample_rate=44100,
        normalize_local_scores=False):
    """
    Function that applies isolation technique determined by
    isolation_parameters dictionary across a folder of audio clips.

    Args:
        audio_dir (string)
            - Directory with wav audio files.

        isolation_parameters (dict)
            - Python Dictionary that controls the various label creation
              techniques. The only unique key to TweetyNET is tweety_output:
              - tweety_output (bool)
                - Set whether or not to use TweetyNET's original output. 
                - If set to `False`, TweetyNET will use the specified `technique` parameter.
                - default: True

        manual_id (string)
            - controls the name of the class written to the pandas dataframe.

        weight_path (string)
            - File path of weights to be used by TweetyNet for
              determining presence of bird sounds.

        normalized_sample_rate (int)
            - Sampling rate that the audio files should all be normalized to.

        normalize_local_scores (bool) # may want to incorporate into isolation parameters
            - Flag to normalize the local scores.

    Returns:
        Dataframe of automated labels for the audio clips in audio_dir.
    """

    # init detector
    device = torch.device('cpu')
    detector = TweetyNetModel(2, (1, 86, 86), 86, device)

    # init labels dataframe
    annotations = pd.DataFrame()
    # generate local scores for every bird file in chosen directory
    for audio_file in os.listdir(audio_dir):
        # skip directories
        if os.path.isdir(audio_dir + audio_file):
            continue

        # Reading in the audio files using librosa, converting to single channeled data with original sample rate
        # Reason for the factor for the signal is explained here: https://stackoverflow.com/questions/53462062/pyaudio-bytes-data-to-librosa-floating-point-time-series
        # Librosa scales down to [-1, 1], but the models require the range [-32768, 32767], so the multiplication is required
        try:
            SIGNAL, SAMPLE_RATE = librosa.load(audio_dir + audio_file, sr=None, mono=True)
            SIGNAL = SIGNAL * 32768
        except BaseException:
            print("Failed to load", audio_file)
            continue

        # downsample the audio if the sample rate isn't 44.1 kHz
        # Force everything into the human hearing range.
        # May consider reworking this function so that it upsamples as well
        try:
            if SAMPLE_RATE != normalized_sample_rate:
                rate_ratio = normalized_sample_rate / SAMPLE_RATE
                SIGNAL = scipy_signal.resample(
                    SIGNAL, int(len(SIGNAL) * rate_ratio))
                SAMPLE_RATE = normalized_sample_rate
        except:
            print("Failed to Downsample" + audio_file)

        # convert stereo to mono if needed
        # Might want to compare to just taking the first set of data.
        if len(SIGNAL.shape) == 2:
            SIGNAL = SIGNAL.sum(axis=1) / 2
        # detection
        try:
            tweetynet_features = compute_features([SIGNAL])
            predictions, local_scores = detector.predict(tweetynet_features, model_weights=weight_path, norm=normalize_local_scores)
        except BaseException as e:
            print("Error in detection, skipping", audio_file)
            print(e)
            continue

        try:
            # Running moment to moment algorithm and appending to a master
            # dataframe. 
            if isolation_parameters["tweety_output"]:
                new_entry = predictions_to_kaleidoscope(
                    predictions, 
                    SIGNAL, 
                    audio_dir, 
                    audio_file, 
                    manual_id, 
                    SAMPLE_RATE)
            else:
                new_entry = isolate(
                    local_scores[0],
                    SIGNAL,
                    SAMPLE_RATE,
                    audio_dir,
                    audio_file,
                    isolation_parameters,
                    manual_id=manual_id,
                    normalize_local_scores=normalize_local_scores)
            # print(new_entry)
            if annotations.empty:
                annotations = new_entry
            else:
                annotations = annotations.append(new_entry)
        except BaseException as e:
            print("Error in isolating bird calls from", audio_file)
            print(e)
            continue
    # Quick fix to indexing
    annotations.reset_index(inplace=True, drop=True)
    return annotations


def generate_automated_labels(
        audio_dir,
        isolation_parameters,
        manual_id="bird",
        weight_path=None,
        normalized_sample_rate=44100,
        normalize_local_scores=False):
    """
    Function that generates the bird labels across a folder of audio clips
    given the isolation_parameters

    Args:
        audio_dir (string)
            - Directory with wav audio files.

        isolation_parameters (dict)
            - Python Dictionary that controls the various label creation
              techniques.

        manual_id (string)
            - controls the name of the class written to the pandas dataframe

        weight_path (string)
            - File path of weights to be used by the model for
              determining presence of bird sounds.

        normalized_sample_rate (int)
            - Sampling rate that the audio files should all be normalized to.
              Used only for the Microfaune model.
        
        normalize_local_scores (bool)
            - Set whether or not to normalize the local scores.

    Returns:
        Dataframe of automated labels for the audio clips in audio_dir.
    """

    #try:
    if(isolation_parameters["model"] == 'microfaune'):
        annotations = generate_automated_labels_microfaune(
                        audio_dir=audio_dir,
                        isolation_parameters=isolation_parameters,
                        manual_id=manual_id,
                        weight_path=weight_path,
                        normalized_sample_rate=normalized_sample_rate,
                        normalize_local_scores=normalize_local_scores)
    elif(isolation_parameters["model"] == 'birdnet'):
        # We need to delete the some keys from the isolation_parameters
        # because we are unpacking the other arguments
        birdnet_parameters = deepcopy(isolation_parameters)
        keys_to_delete = ['model', 'technique', 'threshold_type',
            'threshold_const', 'chunk_size']
        for key in keys_to_delete:
            birdnet_parameters.pop(key, None)
        annotations = generate_automated_labels_birdnet(
                        audio_dir, birdnet_parameters)
    elif(isolation_parameters['model'] == 'tweetynet'):
        annotations = generate_automated_labels_tweetynet(
                        audio_dir=audio_dir,
                        isolation_parameters=isolation_parameters,
                        manual_id=manual_id,
                        weight_path=weight_path,
                        normalized_sample_rate=normalized_sample_rate,
                        normalize_local_scores=normalize_local_scores)
    else:
        print("{model_name} model does not exist"\
            .format(model_name=isolation_parameters["model"]))
    # except:
    #     print("Error. Check your isolation_parameters")
    #     return None
    return annotations

def kaleidoscope_conversion(df):
    """
    Function that strips away Pandas Dataframe columns necessary for PyHa
    package that aren't compatible with Kaleidoscope software

    Args:
        df (Pandas Dataframe)
            - Dataframe compatible with PyHa package whether it be human labels
              or automated labels.

    Returns:
        Pandas Dataframe compatible with Kaleidoscope.
    """
    kaleidoscope_df = [df["FOLDER"].str.rstrip("/\\"), df["IN FILE"], df["CHANNEL"],
                       df["OFFSET"], df["DURATION"], df["MANUAL ID"]]
    headers = ["FOLDER", "IN FILE", "CHANNEL",
               "OFFSET", "DURATION", "MANUAL ID"]

    kaleidoscope_df = pd.concat(kaleidoscope_df, axis=1, keys=headers)
    return kaleidoscope_df

# def annotation_combiner(df):
#    # Initializing the output Pandas dataframe
#    combined_annotation_df = pd.DataFrame()
    # looping through each annotation in the passed in dataframe
#    for annotation in df.index:
    # the case for the first iteration.
#        if combined_annotation_df.empty:
#            combined_annotation_df = df.loc[annotation,:]
#        else:
#            combined_annotation_df = combined_annotation_df.append()
# keeps track of how many annotations have been added to the current
# annotation.
#        annotation_chain_count = 0
# Boolean to keep track whether or not an annotation should be combined with
# the current annotation
#        chain_break = False
    # keeping track of where the current annotation starts
#        cur_offset = df.loc[annotation,"OFFSET"]
#        cur_duration = df.loc[annotation,"DURATION"]
#        start_offset = cur_offset+cur_duration
#        while chain_break == False:
#            annotation_chain_count = annotation_chain_count + 1
#            next_offset = df.loc[annotation+annotation_chain_count,"OFFSET"]
#            next_duration df.loc[annotation+annotation_chain_count,"DURATION"]
    # case in which an annotation overlaps
#            if next_offset <= start_offset:


#        annotation = annotation + annotation_chain_count - 1



from .microfaune_package.microfaune.detection import RNNDetector
from .microfaune_package.microfaune import audio
from .tweetynet_package.tweetynet.TweetyNetModel import TweetyNetModel
from .tweetynet_package.tweetynet.Load_data_functions import compute_features
import torch
import librosa
import matplotlib.pyplot as plt
import pandas as pd
import scipy.signal as scipy_signal
import numpy as np
import seaborn as sns
from .IsoAutio import *

def spectrogram_graph(
        clip_name,
        sample_rate,
        samples,
        automated_df=None,
        premade_annotations_df=None,
        premade_annotations_label="Human Labels",
        save_fig=False):
    """
    Function that produces graphs with the spectrogram of an audio
    clip. Now integrated with Pandas so you can visualize human and
    automated annotations.

    Args:
        clip_name (string)
            - Directory of the clip.

        sample_rate (int)
            - Sample rate of the audio clip, usually 44100.

        samples (list of ints)
            - Each of the samples from the audio clip.

        automated_df (Dataframe)
            - Dataframe of automated labelling of the clip.

        premade_annotations_df (Dataframe)
            - Dataframe labels that have been made outside of the scope of this
              function.

        premade_annotations_label (string)
            - Descriptor of premade_annotations_df

        save_fig (boolean)
            - Whether the clip should be saved in a directory as a png file.

    Returns:
        None
    """
    # Calculating the length of the audio clip
    duration = samples.shape[0] / sample_rate
    time_stamps = np.arange(0, duration, step=1)

    # general graph features
    fig, axs = plt.subplots(1)
    fig.set_figwidth(22)
    fig.set_figheight(5)
    fig.suptitle("Spectrogram for " + clip_name)

    # spectrogram plot
    # Will require the input of a pandas dataframe
    Pxx, freqs, bins, im = axs.specgram(
                                            samples,
                                            Fs=sample_rate,
                                            NFFT=4096,
                                            noverlap=2048,
                                            window=np.hanning(4096),
                                            cmap="ocean")
    axs.set_xlim(0, duration)
    axs.set_ylim(0, 22050)
    axs.grid(which='major', linestyle='-')

    # if automated_df is not None:
    if not automated_df.empty:
        ndx = 0
        for row in automated_df.index:
            minval = automated_df["OFFSET"][row]
            maxval = automated_df["OFFSET"][row] + \
                automated_df["DURATION"][row]
            axs.axvspan(xmin=minval, xmax=maxval, facecolor="yellow",
                           alpha=0.4, label="_" * ndx + "Automated Labels")
            ndx += 1

    # Adding in the optional premade annotations from a Pandas DataFrame
    if not premade_annotations_df.empty:
        ndx = 0
        for row in premade_annotations_df.index:
            minval = premade_annotations_df["OFFSET"][row]
            maxval = premade_annotations_df["OFFSET"][row] + \
                premade_annotations_df["DURATION"][row]
            axs.axvspan(
                xmin=minval,
                xmax=maxval,
                facecolor="red",
                alpha=0.4,
                label="_" *
                ndx +
                premade_annotations_label)
            ndx += 1
    axs.legend()

    # save graph
    if save_fig:
        plt.savefig(clip_name + "_Local_Score_Graph.png")

def local_line_graph(
        local_scores,
        clip_name,
        sample_rate,
        samples,
        automated_df=None,
        premade_annotations_df=None,
        premade_annotations_label="Human Labels",
        log_scale=False,
        save_fig=False,
        normalize_local_scores=False):
    """
    Function that produces graphs with the local score plot and spectrogram of
    an audio clip. Now integrated with Pandas so you can visualize human and
    automated annotations.

    Args:
        local_scores (list of floats)
            - Local scores for the clip determined by the RNN.

        clip_name (string)
            - Directory of the clip.

        sample_rate (int)
            - Sample rate of the audio clip, usually 44100.

        samples (list of ints)
            - Each of the samples from the audio clip.

        automated_df (Dataframe)
            - Dataframe of automated labelling of the clip.

        premade_annotations_df (Dataframe)
            - Dataframe labels that have been made outside of the scope of this
              function.

        premade_annotations_label (string)
            - Descriptor of premade_annotations_df

        log_scale (boolean)
            - Whether the axis for local scores should be logarithmically
              scaled on the plot.

        save_fig (boolean)
            - Whether the clip should be saved in a directory as a png file.

    Returns:
        None
    """
    # Calculating the length of the audio clip
    duration = samples.shape[0] / sample_rate
    # Calculating the number of local scores outputted by Microfaune
    num_scores = len(local_scores)
    # the case for normalizing the local scores between [0,1]
    if normalize_local_scores:
        local_scores_max = max(local_scores)
        for ndx in range(num_scores):
            local_scores[ndx] = local_scores[ndx] / local_scores_max
    # Making sure that the local score of the x-axis are the same across the
    # spectrogram and the local score plot
    step = duration / num_scores
    time_stamps = np.arange(0, duration, step)

    if len(time_stamps) > len(local_scores):
        time_stamps = time_stamps[:-1]

    # general graph features
    fig, axs = plt.subplots(2)
    fig.set_figwidth(22)
    fig.set_figheight(10)
    fig.suptitle("Spectrogram and Local Scores for " + clip_name)
    # score line plot - top plot
    axs[0].plot(time_stamps, local_scores)
    #Look into this and their relation.
    axs[0].set_xlim(0, duration)
    if log_scale:
        axs[0].set_yscale('log')
    else:
        axs[0].set_ylim(0, 1)
    axs[0].grid(which='major', linestyle='-')
    # Adding in the optional automated labels from a Pandas DataFrame
    # if automated_df is not None:
    if not automated_df.empty:
        ndx = 0
        for row in automated_df.index:
            minval = automated_df["OFFSET"][row]
            maxval = automated_df["OFFSET"][row] + \
                automated_df["DURATION"][row]
            axs[0].axvspan(xmin=minval, xmax=maxval, facecolor="yellow",
                           alpha=0.4, label="_" * ndx + "Automated Labels")
            ndx += 1
    # Adding in the optional premade annotations from a Pandas DataFrame
    if not premade_annotations_df.empty:
        ndx = 0
        for row in premade_annotations_df.index:
            minval = premade_annotations_df["OFFSET"][row]
            maxval = premade_annotations_df["OFFSET"][row] + \
                premade_annotations_df["DURATION"][row]
            axs[0].axvspan(
                xmin=minval,
                xmax=maxval,
                facecolor="red",
                alpha=0.4,
                label="_" *
                ndx +
                premade_annotations_label)
            ndx += 1
    axs[0].legend()

    # spectrogram - bottom plot
    # Will require the input of a pandas dataframe
    Pxx, freqs, bins, im = axs[1].specgram(
                                            samples,
                                            Fs=sample_rate,
                                            NFFT=4096,
                                            noverlap=2048,
                                            window=np.hanning(4096),
                                            cmap="ocean")
    axs[1].set_xlim(0, duration)
    axs[1].set_ylim(0, 22050)
    axs[1].grid(which='major', linestyle='-')

    # save graph
    if save_fig:
        plt.savefig(clip_name + "_Local_Score_Graph.png")

# TODO rework function so that instead of generating the automated labels, it
# takes the automated_df as input same as it does with the manual dataframe.

def spectrogram_visualization(
        clip_path,
        weight_path=None,
        premade_annotations_df=None,
        premade_annotations_label="Human Labels",
        automated_df=None,
        isolation_parameters=None,
        log_scale=False,
        save_fig=False,
        normalize_local_scores=False):
    """
    Wrapper function for the local_line_graph and spectrogram_graph functions
    for ease of use. Processes clip for local scores to be used for the
    local_line_graph function.

    Args:
        clip_path (string)
            - Path to an audio clip.

        weight_path (string)
            - Weights to be used for RNNDetector.

        premade_annotations_df (Dataframe)
            - Dataframe of annotations to be displayed that have been created
              outside of the function.

        premade_annotations_label (string)
            - String that serves as the descriptor for the premade_annotations
              dataframe.

        automated_df (Dataframe)
            - Whether the audio clip should be labelled by the isolate function
              and subsequently plotted.

        log_scale (boolean)
            - Whether the axis for local scores should be logarithmically
              scaled on the plot.

        save_fig (boolean)
            - Whether the plots should be saved in a directory as a png file.

    Returns:
        None
    """

    # Reading in the audio file using librosa, converting to single channeled data with original sample rate
    # Reason for the factor for the signal is explained here: https://stackoverflow.com/questions/53462062/pyaudio-bytes-data-to-librosa-floating-point-time-series
    # Librosa scales down to [-1, 1], but the models require the range [-32768, 32767], so the multiplication is required
    try:
        SIGNAL, SAMPLE_RATE = librosa.load(clip_path, sr=None, mono=True)
        SIGNAL = SIGNAL * 32768
    except BaseException:
        print("Failure in loading", clip_path)
        return
    # Downsample the audio if the sample rate > 44.1 kHz
    # Force everything into the human hearing range.
    try:
        if SAMPLE_RATE > 44100:
            rate_ratio = 44100 / SAMPLE_RATE
            SIGNAL = scipy_signal.resample(
                SIGNAL, int(len(SIGNAL) * rate_ratio))
            SAMPLE_RATE = 44100
    except BaseException:
        print("Failure in downsampling", clip_path)
        return

    # Converting to Mono if Necessary
    if len(SIGNAL.shape) == 2:
        # averaging the two channels together
        SIGNAL = SIGNAL.sum(axis=1) / 2

    # Generate parameters for specific models
    local_scores = None
    if(isolation_parameters is not None):
        if(isolation_parameters["model"] == 'microfaune'):
            # Initializing the detector to baseline or with retrained weights
            if weight_path is None:
                # Microfaune RNNDetector class
                detector = RNNDetector()
            else:
                try:
                    # Initializing Microfaune hybrid CNN-RNN with new weights
                    detector = RNNDetector(weight_path)
                except BaseException:
                    print("Error in weight path:", weight_path)
                    return
            try:
                # Computing Mel Spectrogram of the audio clip
                microfaune_features = detector.compute_features([SIGNAL])
                # Running the Mel Spectrogram through the RNN
                global_score, local_score = detector.predict(microfaune_features)
                local_scores = local_score[0].tolist()
            except BaseException:
                print(
                    "Skipping " +
                    clip_path +
                    " due to error in Microfaune Prediction")
        elif (isolation_parameters["model"] == 'tweetynet'):
            # Initializing the detector to baseline or with retrained weights
            device = torch.device('cpu')
            detector = TweetyNetModel(2, (1, 86, 86), 86, device, binary = False)

            try:
                #need a function to convert a signal into a spectrogram and then window it
                tweetynet_features = compute_features([SIGNAL])
                predictions, local_score = detector.predict(tweetynet_features, model_weights=weight_path)
                local_scores = local_score[0].tolist()
            except BaseException:
                print(
                    "Skipping " +
                    clip_path +
                    " due to error in TweetyNet Prediction")
                return None

    # In the case where the user wants to look at automated bird labels
    if premade_annotations_df is None:
            premade_annotations_df = pd.DataFrame()

    # Generate labels based on the model
    if (automated_df is not None):
        if (isinstance(automated_df, bool) and not automated_df):
            automated_df = pd.DataFrame()
            pass
        # Check if Microfaune or TweetyNET was used to generate local scores
        if (local_scores is not None):
            # TweetyNET techniques and output
            if (isolation_parameters["model"] == "tweetynet"
                and isolation_parameters["tweety_output"]):
                automated_df = predictions_to_kaleidoscope(
                                predictions, 
                                SIGNAL, 
                                "Doesn't", 
                                "Doesn't", 
                                "Matter", 
                                SAMPLE_RATE)
            # Isolation techniques
            else: 
                automated_df = isolate(
                        local_score[0],
                        SIGNAL,
                        SAMPLE_RATE,
                        audio_dir = "",
                        filename = "",
                        isolation_parameters=isolation_parameters)
        # Catch, generate the labels for other models
        else:
            automated_df = generate_automated_labels(
                audio_dir=clip_path,
                isolation_parameters=isolation_parameters,
                weight_path=weight_path,
                normalized_sample_rate=SAMPLE_RATE,
                normalize_local_scores=normalize_local_scores)

        if (len(automated_df["IN FILE"].unique()) > 1):
            print("\nWarning: This function only generates spectrograms for one clip. " +
                  "automated_df has annotations for more than one clip.")
    else:
        automated_df = pd.DataFrame()

    # If local scores were generated, plot them AND spectrogram
    if (local_scores is not None):
        local_line_graph(
                local_scores,
                clip_path,
                SAMPLE_RATE,
                SIGNAL,
                automated_df,
                premade_annotations_df,
                premade_annotations_label=premade_annotations_label,
                log_scale=log_scale,
                save_fig=save_fig,
                normalize_local_scores=normalize_local_scores)
    else:
        spectrogram_graph(
            clip_path,
            SAMPLE_RATE,
            SIGNAL,
            automated_df=automated_df,
            premade_annotations_df=premade_annotations_df,
            premade_annotations_label=premade_annotations_label,
            save_fig=save_fig)

def binary_visualization(automated_df, human_df, save_fig=False):
    """
    Function to visualize automated and human annotation scores across an audio
    clip in the form of multiple binary plots that represent different metrics.

    Args:
        automated_df (Dataframe)
            - Dataframe of automated labels for one clip

        human_df (Dataframe)
            - Dataframe of human labels for one clip.

        plot_fig (boolean)
            - Whether or not the efficiency statistics should be displayed.

        save_fig (boolean)
            - Whether or not the plot should be saved within a file.

    Returns:
        Dataframe with statistics comparing the automated and human labeling.
    """
    duration = automated_df["CLIP LENGTH"].to_list()[0]
    SAMPLE_RATE = automated_df["SAMPLE RATE"].to_list()[0]
    # Initializing two arrays that will represent the
    # labels with respect to the audio clip
    # print(SIGNAL.shape)
    human_arr = np.zeros((int(SAMPLE_RATE * duration),))
    bot_arr = np.zeros((int(SAMPLE_RATE * duration),))

    folder_name = automated_df["FOLDER"].to_list()[0]
    clip_name = automated_df["IN FILE"].to_list()[0]
    # Placing 1s wherever the au
    for row in automated_df.index:
        minval = int(round(automated_df["OFFSET"][row] * SAMPLE_RATE, 0))
        maxval = int(
            round(
                (automated_df["OFFSET"][row] +
                 automated_df["DURATION"][row]) *
                SAMPLE_RATE,
                0))
        bot_arr[minval:maxval] = 1
    for row in human_df.index:

        minval = int(round(human_df["OFFSET"][row] * SAMPLE_RATE, 0))
        maxval = int(
            round(
                (human_df["OFFSET"][row] +
                 human_df["DURATION"][row]) *
                SAMPLE_RATE,
                0))
        human_arr[minval:maxval] = 1

    human_arr_flipped = 1 - human_arr
    bot_arr_flipped = 1 - bot_arr

    true_positive_arr = human_arr * bot_arr
    false_negative_arr = human_arr * bot_arr_flipped
    false_positive_arr = human_arr_flipped * bot_arr
    true_negative_arr = human_arr_flipped * bot_arr_flipped
    IoU_arr = human_arr + bot_arr
    IoU_arr[IoU_arr == 2] = 1

    plt.figure(figsize=(22, 10))
    plt.subplot(7, 1, 1)
    plt.plot(human_arr)
    plt.title("Ground Truth for " + clip_name)
    plt.subplot(7, 1, 2)
    plt.plot(bot_arr)
    plt.title("Automated Label for " + clip_name)

    # Visualizing True Positives for the Automated Labeling
    plt.subplot(7, 1, 3)
    plt.plot(true_positive_arr)
    plt.title("True Positive for " + clip_name)

    # Visualizing False Negatives for the Automated Labeling
    plt.subplot(7, 1, 4)
    plt.plot(false_negative_arr)
    plt.title("False Negative for " + clip_name)

    plt.subplot(7, 1, 5)
    plt.plot(false_positive_arr)
    plt.title("False Positive for " + clip_name)

    plt.subplot(7, 1, 6)
    plt.plot(true_negative_arr)
    plt.title("True Negative for " + clip_name)

    plt.subplot(7, 1, 7)
    plt.plot(IoU_arr)
    plt.title("Union for " + clip_name)

    plt.tight_layout()
    if save_fig:
        x = clip_name.split(".")
        clip_name = x[0]
        plt.save_fig(clip_name + "_label_plot.png")

def annotation_duration_histogram(
    annotation_df,
    n_bins = 6,
    min_length = None,
    max_length = None,
    save_fig = False,
    title = "Annotation Length Histogram",
    filename = "annotation_histogram.png"):
    """
    Function to build a histogram so a user can visually see the length of
    the annotations they are working with.

    Args:
        annotation_df (Dataframe)
            - Dataframe of automated or human labels

        n_bins (int)
            - number of histogram bins in the final histogram
            - default: 6

        min_length (int)
            - minimum length of the audio clip
            - default: 0s

        max_length (int)
            - maximum length of the audio clip
            - default: 60s

        save_fig (boolean)
            - Whether or not the histogram should be saved as a file.
            - default: False

        filename (string)
            - Name of the file to save the histogram to.
            - default: "annotation_histogram.png"

    Returns:
        Histogram of the length of the annotations.
    """
    # Create the initial histogram
    duration = annotation_df["DURATION"].to_list()
    fig, ax = plt.subplots()
    sns_hist = sns.histplot(
        data=duration,
        bins=n_bins,
        line_kws=dict(edgecolor="k", linewidth=2),
        stat="count",
        ax=ax)

    # Modify the length of the x-axis as specified
    if max_length is not None and min_length is not None:
        if max_length < min_length:
            raise ValueError("max_length cannot be less than `min_length")
        plt.xlim(min_length, max_length)
    elif max_length is not None:
        plt.xlim(right=max_length)
    elif min_length is not None:
        plt.xlim(left=min_length)

    # Set title and the labels
    sns_hist.set_title(title)
    sns_hist.set(xlabel="Annotation Length (s)", ylabel = "Count")

    # Save the histogram if specified
    if save_fig:
        sns_hist.get_figure().savefig(filename)



from setuptools import setup, find_packages
from pydoc import ModuleScanner 


modules = ["Keras", "librosa", "matplotlib", "microfaune-ai", "numpy",
    "opensoundscape", "pandas", "pydub", "scikit-image", "scikit-learn",
    "scipy", "seaborn", "SoundFile", "tensorflow", "torch", "freetype-py"
]
#blacklist_mod = ["multiprocessing", 'raystreaming', 'formatter' ]
#def callback(path, modname, desc, modules=modules):
#    if modname and modname[-9:] == '.__init__':
#        modname = modname[:-9] + ' (package)'
#    if modname.find('.') < 0 and modname[0] != '_'and modname[0] != '~' and not(modname in blacklist_mod):
#        modules.append(modname)
#def onerror(modname):
#    callback(None, modname, None)#

#ModuleScanner().run(callback, onerror=onerror)

print(modules)

# Setting up
setup(
       # the name must match the folder name 'verysimplemodule'
        name="PyHa_test", 
        version="0.0.19",
        author="Jacob Ayers, Sean Perry, Sam Prestrelski, Vannessa Salgado",
        #author_email="<youremail@email.com>",
        description="A python package for automatically detecting species and comparing to ground truth",
        #long_description=LONG_DESCRIPTION,
        packages=find_packages(),
        install_requires=modules, # add any additional packages that 
        # needs to be installed along with your package. Eg: 'caer'
        
        keywords=['python'],
        classifiers= [
            "Development Status :: 3 - Alpha",
            "Programming Language :: Python :: 3.7",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: Microsoft :: Windows",
        ]
)